/*
 * lodowka_fw.c
 *
 * Created: 14.04.2024 13:01:29
 * Author : Grzegorz Mazurewicz
 *
 * fuses: High 0xD1, Low 0x04 (8MHz internal)
 */
#define F_CPU 8000000
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

// ZMIENNE GLOBALNE
volatile static uint8_t timer488Hz;     //timer 488Hz  od 0 do 255
volatile static uint16_t led_number = 0;  //wyswietlana liczba na ekranie
//0-999 - liczba w formacie 99.9, 1000-1999 liczba w formacie 99c
static uint8_t led_digit = 0;       //aktualnie wyswietlana cyfra 2,1,0,2,1,0,.... (bez volatile, uzywana tylko w przerwaniu)
static uint16_t current_temp, target_temp; //temperatury w formacie fixed 0.00
static uint16_t temp_tab[4]; //tablica pomiarow
static uint8_t temp_tab_nr = 0; //nr aktualnego pomiaru do wykonania

//////////////////////////////////////////////////////////////////////////
//  OBSLUGA PRZERWAN I WYSWIETLACZA
//

//PB0-nc, PB1-nc, PB2-LEDA, PB3-mosi, PB4-miso, PB5-sck
//PC0-LEDF, PC1-BTN_R, PC2-LEDE, PC3-LEDD, PC4-LEDDP, PC5-LEDG
//PD0-LEDC, PD1-LEDB, PD2-BTN_L, PD3-PELTIER, PD4-TEMP_SENSOR, PD5-DIG3, PD6-DIG2, PD7-DIG1

//maska wlaczania segmentow cyfr dla cyfr od 0 do 9, stopnia i pustego znaku
const static uint8_t ledmaskpb[12] = {0b11111011, 0b11111111, 0b11111011, 0b11111011, 0b11111111, 0b11111011,
                                      0b11111011, 0b11111011, 0b11111011, 0b11111011, 0b11111011, 0b11111111
                                     };
const static uint8_t ledmaskpc[12] = {0b11110010, 0b11111111, 0b11010011, 0b11010111, 0b11011110, 0b11010110,
                                      0b11010010, 0b11111111, 0b11010010, 0b11010110, 0b11011110, 0b11111111
                                     };
const static uint8_t ledmaskpd[12] = {0b11111100, 0b11111100, 0b11111101, 0b11111100, 0b11111100, 0b11111110,
                                      0b11111110, 0b11111100, 0b11111100, 0b11111100, 0b11111101, 0b11111111
                                     };
const static uint8_t segmaskpd[3] = {0b11011111, 0b10111111, 0b01111111};

//przerwanie timera
ISR(TIMER0_OVF_vect)
{
  uint8_t led_seg;
  uint16_t led_n;

  if (timer488Hz != 0) timer488Hz--;
  //OBSLUGA WYSWIEETLACZA
  //wygaszenie wszystkiego
  PORTB |= 0b00000100; //wygas wszystkie segmenty
  PORTC |= 0b00111101; //wygas wszystkie segmenty
  PORTD |= 0b11100011; //wygas wszystkie cyfry (logika ujemna) i segmenty w porcie D
  //wygaszenie wyswietlacza gdy liczba jest rowna 0
  if (led_number == 0) return;
  //obliczanie cyfry
  led_digit = (led_digit == 0) ? 2 : led_digit - 1; //zmiana nr cyfry 2,1,0,2,1,0,....
  led_n = (led_number > 1000) ? led_number - 1000 : led_number; //oblicz numer bez 1000 oznaczajcego inny format
  if (led_digit == 2) led_seg = (uint8_t)(led_n / 100 % 10); //oblicz odpowiednia cyfre
  else  if (led_digit == 1) led_seg = (uint8_t)(led_n / 10 % 10);
  else  led_seg = (uint8_t)(led_n  % 10);
  //jezeli mamy tryb pokazywania stopni i jestesmy w ostatniej cyfrze
  if ((led_number >= 1000) && (led_digit == 0)) led_seg = 10;
  //jezeli nie mamy trybu pokazywania stopni i jestesmy w srodkowej cyfrze to dodaj przecinek
  if ((led_number < 1000) && (led_digit == 1)) PORTC &= 0b11101111;
  //jezeli jestesmy w pierwszej cyfrze i jest ona rowna 0 to ja wygaszamy
  if ((led_digit == 2) && (led_seg == 0)) led_seg = 11;
  //ustaw segmenty wyswietlacza
  PORTB &= ledmaskpb[led_seg];
  PORTC &= ledmaskpc[led_seg];
  PORTD &= ledmaskpd[led_seg];
  //wlacz wlasciwy segment
  PORTD &= segmaskpd[led_digit];
}

//////////////////////////////////////////////////////////////////////////
//  OBSLUGA 1 WIRE POMIAR TEMPERATURY
//

//wykonuje reset 1wire
void Reset1Wire(void)
{
  DDRD |= 0b00010000; //wyslij 0 na wyjscie
  _delay_us(500);
  DDRD &= 0b11101111; //przelacz na wejscie
  //tu moznaby sprawdzic czy scalak odpowiedzial
  _delay_us(500);
}

void Send1(void)
{
  //wyslij 0 przez 1-15us, wysoka imp. przez 60us liczone od poczatku, caly bit min 60us ale nie wiecej niz 120us
  DDRD |= 0b00010000;
  _delay_us(5);
  DDRD &= 0b11101111;
  _delay_us(65);
}

void Send0(void)
{
  //wyslij 0 przez 60-120us, pozniej wysoka imp.. caly bit min 60us ale nie wiecej niz 120us
  DDRD |= 0b00010000;
  _delay_us(65);
  DDRD &= 0b11101111;
  _delay_us(5);
}

void SendByte(const uint8_t b)
{
  for (uint8_t a = 1; a != 0; a <<= 1)
  {
    if ((b & a))  Send1();
    else Send0();
  }
}

uint8_t GetBit(void)
{
  uint8_t wynik;
  //wyslij 0 (powyzej 1us), rozpocznij okno po 1-15us , zakoncz okno probkowania po 15us liczac od poczatku, calosc od 60 do 120us
  DDRD |= 0b00010000;
  _delay_us(2);
  DDRD &= 0b11101111; //bez zbednej zwloki bo ma byc ponizej 1us
  _delay_us(8);
  wynik = PIND & 0b00010000;
  _delay_us(60); //poczekajmy reszte zeby calo okno czasowe trwalo 70us
  if ((wynik)) return 1;
  else return 0;
}

//start pomiaru
void Temp_Start(void)
{
  Reset1Wire();
  SendByte(0xCC); //pomin adresowanie sprzetu
  SendByte(0x44); //rozpocznij pomiar
}

//czy pomiar zakonczony (wartosc 1) lub nadal trwa (wartosc 0)
uint8_t Temp_Busy(void)
{
  return GetBit();
}

//odczytuje wynik pomiaru
uint16_t Temp_GetValue(void)
{
  uint16_t wynik;
  Reset1Wire();
  SendByte(0xCC); //pomin adresowanie sprzetu
  SendByte(0xBE); //komenda odczytu 9 bajtow danych
  //odczytujemy 16 bitow zawierajacych wynik, na reszte lejemy
  wynik = 0;
  for (uint8_t a = 0; a < 16; a++)
    wynik |= ((uint16_t)GetBit() << a);
  //jezeli liczba jest ujemna to niech wynikiem bedzie 0 st.C
  if ((wynik & 0x8000) != 0) wynik = 0;
  //przeliczenie na fixed point 0.00stC
  //nalezy wykonac raw*100/16 -> raw*25/4
  return (wynik * 25) >> 2;
}

//////////////////////////////////////////////////////////////////////////
//  OBSLUGA KLAWIATURY I KONFIGUROWANIA TEMPERATURY
//

//0-brak przyciskow, 1-lewy, 2-prawy, 3-oba przyciski naraz
uint8_t KeyGet(void)
{
  uint8_t button = 0;

  if ((~PINC & 0b00000010)) button = 0x01; //przycisk lewy
  if ((~PIND & 0b00000100)) button |= 0x02; //przycisk prawy
  return button;
}

void KeyWait(void)
{
  uint8_t cnt;

  cnt = 10;
  do //w petli czekaj przez 100ms az na pewno nie bedzie nacisnietego klawisza i zadnych drgan stykow
  {
    if ((KeyGet())) cnt = 10;
    _delay_ms(10);
    cnt--;
  }
  while ((cnt));
}

//////////////////////////////////////////////////////////////////////////
//  GLOWNY KOD PROGRAMU
//

void Atmel_Init(void)
{
//PB0-nc, PB1-nc, PB2-LEDA, PB3-mosi, PB4-miso, PB5-sck
//PC0-LEDF, PC1-BTN_L, PC2-LEDE, PC3-LEDD, PC4-LEDDP, PC5-LEDG
//PD0-LEDC, PD1-LEDB, PD2-BTN_R, PD3-PELTIER, PD4-TEMP_SENSOR, PD5-DIG3, PD6-DIG2, PD7-DIG1
//BTN, DIGx, LEDx - stan aktywny niski, PELTIER stan aktywny wysoki
  DDRB  = 0b00000100; //LEDA-wyjscie
  PORTB = 0b00000100; //LEDA-zgaszony
  DDRC  = 0b00111101; //piny segmentow sa wyjsciowe, BTN wejsciowy
  PORTC = 0b00111111; //domyslnie segmenty zgaszone, BTN ma pullup
  DDRD  = 0b11101011; //piny peltiera, segmentow i cyfr wyjsciowe, 1wire jest wejsciowe
  PORTD = 0b11100111; //wszystko wylaczone, 1wire nie ma pullup, BTN ma pullup
  //timer 0  f=8000000/(64*256) = 488Hz
  TCCR0 = (0 << CS02) | (1 << CS01) | (1 << CS00); //preskaler 1/64
  TIMSK = (1 << TOIE0); //wlacz przerwanie dla timer0
  sei(); //wlaczamy na stale przerwania
}

void Peltier(const uint8_t on)
{
  if ((on)) PORTD |= (1 << PIND3);
  else PORTD &= (~(1 << PIND3));
}

void SetTemp(void)
{
  uint16_t new_temp;
  uint8_t timer_config;

  KeyWait();
  new_temp = target_temp;
  timer_config = 50;
  do
  {
    led_number = 1000 + new_temp / 10;
    if (KeyGet() == 2)
    {
      KeyWait();
      if (new_temp < 2000) new_temp += 100;
      timer_config = 50;
    }
    if (KeyGet() == 1)
    {
      KeyWait();
      if (new_temp > 500) new_temp -= 100;
      timer_config = 50;
    }
    _delay_ms(100);
    timer_config--;
  }
  while (timer_config != 0);
  if (new_temp != target_temp)
  {
    target_temp = new_temp;
    eeprom_write_word(0, target_temp);
  }
}

int main(void)
{
  Atmel_Init();
  //jak najszybciej rozpocznij pierwszy pomiar temperatury
  Temp_Start();
  //odczyt i ewentualna korekta konfiguracji temperatury
  target_temp = eeprom_read_word(0);
  target_temp = target_temp / 100 * 100; //wyzerowanie dwoch ostatnich cyfr
  if (target_temp < 500) target_temp = 500;
  if (target_temp > 2000) target_temp = 2000;
  //ustaw tablice pomiarow na cos neutralnego
  for (uint8_t a = 0; a < 4; a++)
    temp_tab[a] = 2000; //20 stC
  led_number = 200; //wyswietl wstepnie tez 20 stC
  while (1)
  {
    timer488Hz = 244; //odmierzaj 0.5 sekundy
    //zmierz temperature i rozpocznij nowy pomiar
    temp_tab[temp_tab_nr] = Temp_GetValue();
    temp_tab_nr = (temp_tab_nr == 0) ? 3 : temp_tab_nr - 1;
    Temp_Start();
    //policz temperature srednia
    current_temp = (temp_tab[0] + temp_tab[1] + temp_tab[2] + temp_tab[3]) >> 2;
    //pokaz temperature
    led_number = current_temp / 10;
    //sterowanie Peltierem
    if (current_temp >= target_temp) Peltier(1);
    else Peltier(0);
    //poczekaj na koniec pol sekundy
    while ((timer488Hz)) //czekaj na koniec pol sekundy przy okazji sprawdzajac klawiature
      if (KeyGet() != 0) SetTemp(); //wejdz do ustawien
  }
}
